 /*******************************************************************************
	ProjectName: Frequnency Generator Controlle

	FileName:    main.c

  Copyright(c) 2022.8  Suwa-Koubou

   pin assign PIC32MX220F032B

                                 MCLR 1    28 AVDD       ------ 3.3V
          debug to PC <----  U1TX/RA0 2    27 AVSS       ------ GND
                                  RA1 3    26 RB15       -----> Analog Switch B (0.01uF)
Analog Switch A(10uF) <---- PGED1/RB0 4    25 RB14       -----> Analog Switch c (0.1uF)
Analog Switch D(1uF)  <---- PGEC1/RB1 5    24 RB13/T4CK/INT2 <----- XR-2206 Pulse Out
       debug from PC  ---->  U1RX/RB2 6    23 Vusb3v3
rotary encorder CH-A  ---->       RB3 7    22 RB11/D-  
                 GND  -----   GND/VSS 8    21 RB10/D+  
           20MHz    -+-----  OSC1/RA2 9    20 VCAP via 10uF to  GND
           Cerarock  +----- OSC2/RA3 10    19 GND       ------- GND  
rotary encorder CH-B  ---->      RB4 11    18 RB9/SDA1  ------> LCD Display SDA
              Act LED <----      RA4 12    17 RB8/SCL1  ------> LCD Display SCL
                3.3V --------    VDD 13    16 RB7/OC1   ------> Control Voltage(0.02V-3.0V)
rotary encorder Sw    ---->      RB5 14    15 VBUS


      Timer1   1msec interval timer (include delay_ms() function)
      Timer2   output compare_1 16bit time base (CV out)
      Timer3   Frequency Count measuring time base (0.1sec)
      Timer4   32bit Timer PRx frequency and period count 1Hz - 4.2GHz
      Timer5   32bit Timer PRy


 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "HardwareProfile.h"
#include "typedef.h"
#include "timer.h"
#include "oled.h"

#include "debug.h"


/*------------------------------------------------------------------------------
	chip Configuration PIC32MX220F032B
------------------------------------------------------------------------------*/

#include "p32mx220f_hs20_config.h"


/*------------------------------------------------------------------------------
	definition
------------------------------------------------------------------------------*/

#define  USE5KHZPWM
//#define  USE10KHZPWM
//#define  USE20KHZPWM


#define act_led_on()     LATAbits.LATA4=0
#define act_led_off()    LATAbits.LATA4=1

// state of encorder push switch 
#define SW_OFF           0    // switch off
#define SW_LONG_ON       1    // switch on time is longer than 500msec.
#define SW_SHORT_ON      2    // switch on time is shoter than 500msec. 

// frequency range
#define FREQ_VERY_LOW    0
#define FREQ_LOW         1
#define FREQ_HIGH        2
#define FREQ_VERY_HIGH   3

// range structure
typedef struct {
    char *name;          // range name
    int  select;         // select capacitor
    int  fmin;           // range minimum frequency
    int  fmax;           // range maximum frequency
    int  vcmin;          // fmin controlle voltage.(PWM value)
    int  vcmax;          // fmax controlle voltage.(PWM value)
    int  vcstep;         // vc/Hz(LOW range) or Hz/vc(HIGH range)
    int  vcratio;        
} FREQ_INFO;


/*------------------------------------------------------------------------------
	global variables
------------------------------------------------------------------------------*/
#if defined USE5KHZPWM
const FREQ_INFO FreqInfo[4] = {
  {"FREQ_VERY_LOW",  0b1111111111111110, /* RB0 */     10,    100, 6952, 1628,  80, 0, },    // 80=8000/100    vc/Hz
  {"FREQ_LOW",       0b1111111111111101, /* RB1 */    100,   1000, 6890, 1163,   8, 0, },    //  8=8000/1000   vc/Hz
  {"FREQ_HIGH",      0b1011111111111111, /* RB14 */  1000,  10000, 6893, 1041,   5, 4, },    //  5=4*1.25  1.25=10000/8000  Hz/vc   
  {"FREQ_VERY_HIGH", 0b0111111111111111, /* RB15 */ 10000, 100000, 6896,  752,  50, 4, },    // 50=4*12.5  12.5=100000/8000 Hz/vc
};
#elif defined USE10KHZPWM
const FREQ_INFO FreqInfo[4] = {
  {"FREQ_VERY_LOW",  0b1111111111111110, /* RB0 */     10,    100, 3476,  814, 40, 0, },     //  40=4000/100    vc/Hz
  {"FREQ_LOW",       0b1111111111111101, /* RB1 */    100,   1000, 3445,  582,  4, 0, },     //   4=4000/1000   vc/Hz
  {"FREQ_HIGH",      0b1011111111111111, /* RB14 */  1000,  10000, 3446,  520,  5, 2, },     //  5=2*2.5 2.5=10000/4000  Hz/vc   
  {"FREQ_VERY_HIGH", 0b0111111111111111, /* RB15 */ 10000, 100000, 3448,  376, 50, 2, },     // 50=2*25   25=100000/4000 Hz/vc
};
#elif defined USE20KHZPWM
const FREQ_INFO FreqInfo[4] = {
  {"FREQ_VERY_LOW",  0b1111111111111110, /* RB0 */     10,    100, 1738,  407, 20, 0, },     //  20=2000/100    vc/Hz
  {"FREQ_LOW",       0b1111111111111101, /* RB1 */    100,   1000, 1731,  326,  2, 0, },     //   2=2000/1000   vc/Hz
  {"FREQ_HIGH",      0b1011111111111111, /* RB14 */  1000,  10000, 1723,  260,  5, 1, },     //  5=10000/2000  Hz/vc   
  {"FREQ_VERY_HIGH", 0b0111111111111111, /* RB15 */ 10000, 100000, 1724,  188, 50, 1, },     // 50=100000/2000 Hz/vc
};
#endif

int      encorder_count;

char pbuf[32];


/*------------------------------------------------------------------------------
	function prototype
------------------------------------------------------------------------------*/

int   main (void);

void  frequency_order(void);
void  update_freq(int *freq, int count);

void  freq_control(int freq);
void  set_freq_range(int range);
void  freq_low_tune(int range, int freq);
void  freq_high_tune(int range, int freq);

int   measure_frequency(int meas_time);
int   measure_frequency_by_period(void);

void  io_initialize(void);


/*==============================================================================
	main() start here. 
==============================================================================*/

int  main ( void )
{
    io_initialize();
    
    act_led_on();
    putstr("*** Start Frequency Generator Controlle\r\n");

    oled_putstr2(0,0, "**** Start *****");
    oled_putstr2(0,2, " Suwa Koubou    ");
    delay_ms(200);

    frequency_order();
    return 0;
}


/*==============================================================================
    order frequency
==============================================================================*/

void frequency_order(void)
{
    int   prev, sw;
    int   count;
    int   freq;

    prev = encorder_count;
    freq = 10;
    for(;;){

        sprintf(pbuf, "order: %6dHz ", freq);
        oled_putstr2(0, 0, pbuf);

        // check rotary encorder, switch and encorde count
        sw = check_rotary_switch();

        // when switch push on, call frequency control
        if(sw == SW_LONG_ON || sw == SW_SHORT_ON){
            delay_ms(100);
            freq_control(freq);
        }else{
            // update order frequency by encorder value
            count = encorder_count - prev;
            prev = encorder_count;
            update_freq(&freq, count);
        }
    }
}

/*------------------------------------------------------------------------------
    update frequency value
------------------------------------------------------------------------------*/

#define abs(x)   ((x)>0?(x):-(x))

void update_freq(int *freq, int count)
{
    if(abs(count) > 6){
        count *= 100;
    }else if(abs(count) > 3){
        count *= 10;
    }
    (*freq) += count;
    if(*freq < 10)     *freq = 10;   // min 10Hz
    if(*freq > 100000) *freq = 100000;   // max 100KHz
}


/*=========================================================================
    check rotary encorder switch on/off state

    return code: SW_OFF/SW_LONG_ON/SW_SHORT_ON
=========================================================================*/

int check_rotary_switch(void)
{
    int cnt;

    if(PORTBbits.RB5 == 1){
        delay_ms(50);     // delay for encorder cw or ccw count time
        return SW_OFF;
    }

    cnt=0;
    while(PORTBbits.RB5 == 0){
        delay_ms(10);
        cnt++;
    }

    if(cnt <= 2)  return SW_OFF;     // chattering time is shoter than 20ms
    if(cnt >= 50) return SW_LONG_ON; // 500msec over.

    return SW_SHORT_ON;              // shorter than 500msec
} 


/*------------------------------------------------------------------------------
    CH-A RB3 change notice interrupt,  CH-B is RB4

    encorder countup/countdown at the rise edge or fall edge.
             ____      ____      ____
    CH-A ___|    |____|    |____|    |____  RB3 interrupt
               ____      ____      ____
    CH-B _____|    |____|    |____|    |____RB4 level read

            CCW <---      ----> CW
------------------------------------------------------------------------------*/

void __attribute__ (( interrupt(ipl2soft), vector(_CHANGE_NOTICE_VECTOR) )) _CNBInterrupt(void)
{

    if(PORTBbits.RB3){  // rise edge
        if(PORTBbits.RB4 == 0){
            encorder_count++;
        }else{
            encorder_count--;
        }
    }else{               // fall edge
        if(PORTBbits.RB4 == 0){
            encorder_count--;
        }else{
            encorder_count++;
        }
    }

    IFS1bits.CNBIF = 0;
}


/*==============================================================================
    set frequency range (select Capacitor) 
==============================================================================*/

void set_freq_range(int range)
{
    LATB |= 0b1100000000000011;  // RB15,Rb14,RB1,RB0 to HIGH( not connect C) 
    LATB &= FreqInfo[range].select;
}


/*===============================================================================
    frequency controlle
===============================================================================*/

void freq_control(int freq)
{
    int range;

    // decide frequency range
    if(freq < FreqInfo[FREQ_VERY_LOW].fmin)  freq = FreqInfo[FREQ_VERY_LOW].fmin;
    if(freq > FreqInfo[FREQ_VERY_HIGH].fmax) freq = FreqInfo[FREQ_VERY_HIGH].fmax;

    for(range = FREQ_VERY_LOW; range <= FREQ_VERY_HIGH; range++){
        if(freq >= FreqInfo[range].fmin && freq <= FreqInfo[range].fmax) break;
    }

    sprintf(pbuf, "order frequency: %dHz\r\n", freq); putstr(pbuf);
    sprintf(pbuf, "select range: %s\r\n", FreqInfo[range].name); putstr(pbuf);

    // select frequency range
    set_freq_range(range);

    oled_putstr2(0, 2, "** tunning...   ");

    if(range <= FREQ_LOW){
       freq_low_tune(range, freq);
    }else{
       freq_high_tune(range, freq);
    }

    return;
}

/*------------------------------------------------------------------------------
    low frequency tune control
    vcstep: vc/Hz
------------------------------------------------------------------------------*/

void freq_low_tune(int range, int freq)
{
    const FREQ_INFO  *p;

    int vc, vcstep;
    int fval, ferr;
    int times;

    // calculate vc for freq. tunning.
    p      = &FreqInfo[range];
    vc     = ((p->vcmin - p->vcmax)*(p->fmax -freq))/(p->fmax - p->fmin) + p->vcmax; 
    vcstep = p->vcstep;

    // tunning
    for(times=0,freq *= 10; ; times++){
        OC1RS = vc;
        delay_ms(100);

        fval = measure_frequency_by_period();
        sprintf(pbuf, "freq=%d.%dHz(vc:%d)\r\n", fval/10, fval%10, vc); putstr(pbuf);

        if(fval > freq){
            ferr = fval-freq;
            if(times < 10){
                vc += ferr * vcstep / 10;   // ferr(0.1Hz) vcstep(Hz/vc) 
            }else{
                vc++;
            }
        }else if(fval < freq){
            ferr = freq-fval;
            if(times < 10){
                vc -= ferr * vcstep / 10;
            }else{
                vc--;
            }
        }else{
            break;  // freq match
        }

        if(ferr < (10/vcstep)) break;  // about match
    }

    putstr("complete\r\n");

    sprintf(pbuf, "done(%6d.%1dHz)", fval/10, fval%10);
    oled_putstr2(0, 2, pbuf);
    return;
}

/*------------------------------------------------------------------------------
    high frequency tune control
     Hz/vc
------------------------------------------------------------------------------*/

void freq_high_tune(int range, int freq)
{
    const FREQ_INFO  *p;

    int vc, vcstep, vcratio;
    int fval, ferr;

    // calculate vc for freq. tunning.
    p       = &FreqInfo[range];
    vc      = ((p->vcmin - p->vcmax)*(p->fmax -freq))/(p->fmax - p->fmin) + p->vcmax; 
    vcstep  = p->vcstep;
    vcratio = p->vcratio;

    // tunning
    for(;;){
        OC1RS = vc;
        delay_ms(100);

        fval = measure_frequency(50);   // measure 500msec
        sprintf(pbuf, "freq=%dHz(vc:%d)\r\n", fval, vc); putstr(pbuf);

        if(fval > freq){
            ferr = fval-freq;
            vc += (ferr*vcratio/vcstep);
        }else if(fval < freq){
            ferr = freq-fval;
            vc -= (ferr*vcratio/vcstep);
        }else{
            break;  // freq match
        }
        if(ferr*vcratio < vcstep) break;  // about match
    }

    putstr("complete\r\n");

    sprintf(pbuf, "done (%7dHz)", fval);
    oled_putstr2(0, 2, pbuf);
    return;
}


/*==============================================================================
    measuring frequency
==============================================================================*/

static int  MeasTime;    // Frequency Count Time:  MeasTime * 10msec
static int  MeasDone;    // 0: measuring   1: meas done
static int  FreqVal;     // measured frequency value
static int  PeriodVal;   // measured period value
static unsigned int PrevPeriod;

//
//  return value Frequency
//    meas_time <= 100   1sec measure
int measure_frequency(int meas_time)
{
    // set measuring parameter
    MeasTime = meas_time;
    MeasDone=0;

    // setup Timer3 for measure timing 

    // setup Timer4 and Timer5 32bit counter
    T4CKR = 0b0011;        // RB13 is T4CK, Timer4 clock in from RB13
    T4CONbits.TCS = 1;     // clock in from T4CK 
    T4CONbits.TCKPS = 0;   // prescale 1:1
    T4CONbits.TON=1;       // Go frequency count 

    TMR3=0;
    IFS0bits.T3IF = 0;
    IEC0bits.T3IE = 1;
    TMR4 = 0;
    T3CONbits.TON = 1;
    while(MeasDone==0);
    T3CONbits.TON = 0;
    IEC0bits.T3IE = 0;

    return FreqVal * (100/meas_time);
}

void __attribute__ (( interrupt(ipl4soft), vector(_TIMER_3_VECTOR) )) _T3Interrupt(void)
{
    FreqVal = TMR4;   // read Timer4+Timer5    
    if(--MeasTime ==0){
        MeasDone=1;
    }
    IFS0bits.T3IF = 0;
}

//
//  return value frequency * 10 (unit 0.1Hz)
//
int measure_frequency_by_period(void)
{
    // clear measuring parameter
    MeasDone=0;
    PrevPeriod=0;
    PeriodVal=0;
    MeasTime = 4;

    // setup Timer4 and Timer5 32bit system clock counter
    T4CONbits.TCS = 0;         // clock in from peripheral: clock is 40MHz/TCKPS
    T4CONbits.TCKPS = 0b001;   // prescale 1:2  40MHz/2=20MHz
    TMR4=0;
    T4CONbits.TON=1;           // Go 20MHz clock count 

    // INT2 connected XR-2206 pulse out
    INT2R = 0b0011;    // INT2R asigned to RB13(XR-2206 pulse)
    IFS0bits.INT2IF = 0;
    IEC0bits.INT2IE = 1;
    while(MeasDone==0);
    IEC0bits.INT2IE = 0;

    PeriodVal >>= 2;  // avarage MeasTime(4) times 
    FreqVal = ((400000000/PeriodVal)+1)/2; 

    return FreqVal;   // freq*10
}


void __attribute__ (( interrupt(ipl4soft), vector(_EXTERNAL_2_VECTOR) )) _INT2Interrupt(void)
{
    unsigned int val;

    val = TMR4;

    if(PORTBbits.RB13){ // rise edge
        if(PrevPeriod){
            PeriodVal+=(int)(val-PrevPeriod);
            if(--MeasTime==0){
                MeasDone=1;
            }
        }
        PrevPeriod = val;
    } 

    IFS0bits.INT2IF = 0;
    return;
}


/*==============================================================================
	io initialize
==============================================================================*/

void io_initialize(void)
{
    __builtin_disable_interrupts();
    
    // Data Memory SRAM wait states: Default Setting = 1; set it to 0
    BMXCONbits.BMXWSDRM = 0;

    // enable multi-vector interrupts
    INTCONSET = _INTCON_MVEC_MASK;
 
    // enable global interrupts
    __builtin_enable_interrupts();


    //--- setting I/O port ----------------------------------------
    ANSELA=0;
    ANSELB=0;

    LATA = 0x0000; TRISA = 0xffff;	// RA are default input.
    LATB = 0x0000; TRISB = 0xffff;	// RB are default input.

    // *** frequency range select switch control **********
    LATB  = 0b1100000000000010;   // analog switch D4066 control bits are
    TRISB = 0b0011111111111100;   //  RB0,RB1,RB14,RB15. Initial set RB0 to 0(select FREQ_VERY_LOW).

    // *** Power LED **************************************
    LATAbits.LATA4=1;      // act LED turn off
    TRISAbits.TRISA4=0;

    // *** UART1 ******************************************
    RPA0R = 0b0001;         // RA0 is assigned to U1TX
    LATAbits.LATA0 = 1;		// RA0(U1TX) to level high
    TRISAbits.TRISA0 = 0;	// RA0(U1TX) is output

    U1RXR = 0b0100;         // RB2 is assigned to U1RX
    TRISBbits.TRISB2=1;	    // RB2 is input 


    // *** controlle voltage out put **********************
    // setup 16bit Timer2 for OC1
    T2CON = 0;
    T2CONbits.TCKPS = 0b000;   // 1:1  40MHz

    // setup OC1 to PWM range 0 to 4000
    RPB7R = 0b0101;            // RPB7 --> OC1
    OC1CON = 0;
    OC1CONbits.OCTSEL=0;       // use Timer2
  #if defined USE5KHZPWM
    PR2 = 7999;                // 40MHz/8000=5KHz
    OC1R = 4000; 
    OC1RS = 4000;
  #elif defined USE10KHZPWM
    PR2 = 3999;                // 40MHz/4000=10KHz
    OC1R = 2000; 
    OC1RS = 2000;
  #elif defined USE20KHZPWM
    PR2 = 1999;                // 40MHz/2000=20KHz
    OC1R = 1000; 
    OC1RS = 1000;
  #endif
    OC1CONbits.OCM = 0b110;    // PWM

    // Go output 
    OC1CONbits.ON = 1;
    T2CONbits.ON = 1;

    // setup Timer4 and Timer5 32bit counter
    T4CON = 0;
    T5CON = 0;
    T4CONbits.T32 = 1;         // 32bit T4+T5 
    T4CONbits.TCS = 0;         // clock in from peripheral: clock is 40MHz/TCKPS
    T4CONbits.TCKPS = 0;       // prescale 1:1  40MHz/2=20MHz
    IEC0bits.T4IE = 0;         // don't use interrupt
    IEC0bits.T5IE = 0;
    PR4 = 0xffffffff;          // load PR4, PR5 with 32bit
    //T4CONbits.TON=1;           // Go 20MHz clock count 

    // Timer4 capture by Timer3 interrupt ( frequency direct count )
    T3CON = 0;
    T3CONbits.TCKPS = 0b100;  // prescale 1:16  clock: 40MHz/16=2.5MHz
    PR3 = 24999;              // 2.5MHz/100Hz(10msec)=25000
    IPC3bits.T3IP = 4;
    IFS0bits.T3IF = 0;
    IEC0bits.T3IE = 0;

    // Timer4 capture by INT2 interrupt ( measuring 4th period )
    INTCONbits.INT2EP = 1;     // rising edge, XR-2206 Pulse
    IPC2bits.INT2IP = 4;
    IFS0bits.INT2IF = 0;
    IEC0bits.INT2IE = 0;

    // rotary encorder interrupt
    CNCONB = 0;
    CNENBbits.CNIEB3 = 1;
    IPC8bits.CNIP = 2;
    IFS1bits.CNBIF = 0;
    IEC1bits.CNBIE = 1;
    CNCONBbits.ON = 1;

    // *** interval timer (use Timer1) *******************
    timer_init();

    // *** Init UART *************************************
    UART1Init();

    // *** OLED I2C interface enable *********************
    I2C1CON = 0;
    I2C1BRG = 49;  // PBCLK=40MHz(40000KHz), FSCK=400KHz,
                   // (PBCLK/(2*FSCK))-1=(40000/(2*400))-1=49
    I2C1CONbits.DISSLW = 1;
    I2C1CONbits.I2CEN = 1;

    // initialize oled
    oled_init();

}

/*** end of main.c ************************************************************/

